"""This module defines functions for modifying rmaps in various ways,  generally
the transformations required to automate rmap maintenance on the CRDS website.
"""
import os.path
import sys

from crds.core import (log, utils, rmap, cmdline, config)
from crds.core import exceptions as crexc
from crds.core.log import srepr
from crds import (diff, sync, certify, matches)
import crds

# ============================================================================

class NoUseAfterError(ValueError):
    "The specified UseAfter datetime didn't exist in the rmap."

class NoMatchTupleError(ValueError):
    "The specified Match tuple didn't exist in the rmap."

# ============================================================================

def update_derivation(new_path, old_basename=None):
    """Set the 'derived_from' and 'name' header fields of `new_path`.
    This function works for all Mapping classes:  pmap, imap, and rmap.
    """
    new = rmap.fetch_mapping(new_path)
    if old_basename is None:    # assume new is a copy of old, with old's name in header
        derived_from = new.name
    else:
        derived_from = old_basename
    new.header["derived_from"] = str(derived_from)
    new.header["name"] = str(os.path.basename(new_path))
    new.write(new_path)
    return str(derived_from)

# ============================================================================

def rmap_insert_references(old_rmap, new_rmap, inserted_references):
    """Given the full path of starting rmap `old_rmap`,  modify it by inserting
    or replacing all files in `inserted_references` and write out the result to
    `new_rmap`.    If no actions are performed, don't write out `new_rmap`.

    Return new ReferenceMapping named `new_rmap`
    """
    new = old = rmap.fetch_mapping(old_rmap, ignore_checksum=True)
    new.header["derived_from"] = old.basename
    for reference in inserted_references:
        baseref = os.path.basename(reference)
        with log.augment_exception("In reference", srepr(baseref)):
            log.info("Inserting", srepr(baseref), "into", srepr(new.name))
            new = new.insert_reference(reference)
            log.verbose("Writing", srepr(new_rmap))
            new.write(new_rmap)
    formatted = new.format()
    for reference in inserted_references:
        reference = os.path.basename(reference)
        assert reference in formatted, \
            "Rules update failure. " + srepr(reference) + " does not appear in new rmap." \
            "  May be identical match with other submitted references."
    return new

def rmap_insert_references_by_matches(old_rmap, new_rmap, references_headers):
    """Given the full path of starting rmap `old_rmap`,  modify it by inserting
    or replacing all files in dict `references_headers` which maps a reference file basename
    onto a list of headers under which it should be  matched.  Write out the result to
    `new_rmap`.    If no actions are performed, don't write out `new_rmap`.

    Return new ReferenceMapping named `new_rmap`
    """
    new = old = rmap.load_mapping(old_rmap, ignore_checksum=True)
    for baseref, header in references_headers.items():
        with log.augment_exception("In reference", srepr(baseref)):
            log.info("Inserting", srepr(baseref), "into", srepr(old_rmap))
            log.verbose("Inserting", srepr(baseref), "match case", srepr(header), "into", srepr(old_rmap))
            new = new.insert_header_reference(header, baseref)
    new.header["derived_from"] = old.basename
    log.verbose("Writing", srepr(new_rmap))
    new.write(new_rmap)
    formatted = new.format()
    for baseref in references_headers:
        assert baseref in formatted, \
            "Rules update failure. " + srepr(baseref) + " does not appear in new rmap." \
            "  May be identical match with other submitted references."
    return new

# ============================================================================

def rmap_delete_references(old_rmap, new_rmap, deleted_references):
    """Given the full path of starting rmap `old_rmap`,  modify it by deleting
    all files in `deleted_references` and write out the result to
    `new_rmap`.    If no actions are performed, don't write out `new_rmap`.

    Return new ReferenceMapping named `new_rmap`
    """
    new = old = rmap.load_mapping(old_rmap, ignore_checksum=True)
    for reference in deleted_references:
        baseref = os.path.basename(reference)
        log.info("Deleting", srepr(baseref), "from", srepr(new.name))
        with log.augment_exception("In reference", srepr(baseref)):
            new = new.delete(reference)
    new.header["derived_from"] = old.basename
    log.verbose("Writing", srepr(new_rmap))
    new.write(new_rmap)
    formatted = new.format()
    for reference in deleted_references:
        reference = os.path.basename(reference)
        assert reference not in formatted, \
            "Rules update failure.  Deleted " + srepr(reference) + " still appears in new rmap."
    return new

# ============================================================================

def rmap_check_modifications(old_rmap, new_rmap, old_ref, new_ref, expected=("add",)):
    """Check the differences between `old_rmap` and `new_rmap` and make sure they're
    limited to the types listed in `expected`.

    expected should be "add" or "replace".

    Returns as_expected,  True IFF all rmap modifications match `expected`.
    """
    diffs = diff.mapping_diffs(old_rmap, new_rmap)
    as_expected = True
    for difference in diffs:
        actual = diff.diff_action(difference)
        if actual in expected:
            pass   # white-list so it will fail when expected is bogus.
        else:
            log.error("Expected one of", srepr(expected), "but got", srepr(actual),
                      "from change", srepr(difference))
            as_expected = False
    with open(old_rmap) as pfile:
        old_count = len([line for line in pfile.readlines() if os.path.basename(old_ref) in line])
    with open(new_rmap) as pfile:
        new_count = len([line for line in pfile.readlines() if os.path.basename(new_ref) in line])
    if "replace" in expected and old_count != new_count:
        log.error("Replacement COUNT DIFFERENCE replacing", srepr(old_ref), "with",
                  srepr(new_ref), "in", srepr(old_rmap),
                  old_count, "vs.", new_count)
        as_expected = False
    return as_expected

# ============================================================================

def set_rmap_header(rmapping, new_filename, header_key, header_value, *args, **keys):
    """Set the value of `key` in `filename` to `new_value` and rewrite the rmap.
    This is potentially lossy since rewriting the rmap may/will lose comments and
    formatting quirks.
    """
    log.verbose("Setting header value in", srepr(rmapping.basename), "for", srepr(header_key),
                "=", srepr(header_value))
    try:
        rmapping.header[header_key] = eval(header_value)
    except Exception:
        rmapping.header[header_key] = header_value
    rmapping.write(new_filename)

def del_rmap_header(rmapping, new_filename, header_key):
    """Set the value of `key` in `filename` to `new_value` and rewrite the rmap.
    This is potentially lossy since rewriting the rmap may/will lose comments and
    formatting quirks.
    """
    log.verbose("Deleting header value in", srepr(rmapping.basename), "for", srepr(header_key))
    del rmapping.header[header_key]
    rmapping.write(new_filename)

def del_rmap_parameter(rmapping, new_filename, parameter, *args, **keys):
    """Delete `parameter_name` from the parkey item of the `types` of the specified
    `instruments` in `context`.
    """
    log.info("Deleting parameter", repr(parameter), "from",repr(rmapping.basename))
    parkey = rmapping.parkey
    i, j = get_parameter_index(parkey, parameter)
    del_parkey = parkey[:i] +  ((parkey[i][:j] + parkey[i][j+1:]),)  + parkey[i+1:]
    log.verbose("Replacing", srepr(parkey), "with", srepr(del_parkey), "in", srepr(rmapping.basename))
    rmapping.header["parkey"] = del_parkey
    rmapping.selector.delete_match_param(parameter)
    rmapping.write(new_filename)

def get_parameter_index(parkey, parameter_name):
    """Return the 2D parkey index for `parameter_name`:  (which_sub_tuple,  index_in_subtuple)."""
    for i, pars in enumerate(parkey):
        for j, par in enumerate(pars):
            if par.lower() == parameter_name.lower():
                return i, j
    raise crexc.CrdsError("Can't find index for", repr(parameter_name), "in parkey", repr(parkey))

# ============================================================================

def set_rmap_parkey(rmapping, new_filename, parkey, *args, **keys):
    """Set the parkey of `rmapping` to `parkey` and write out to `new_filename`.
    """
    log.info("Setting parkey, removing all references from", srepr(rmapping.basename))
    pktuple = eval(parkey)
    required_keywords = tuple(utils.flatten(pktuple))
    refnames = rmapping.reference_names()
    references_headers = { refname : get_refactoring_header(rmapping.filename, refname, required_keywords)
                           for refname in refnames }
    rmapping = rmap_delete_references(rmapping.filename, new_filename, refnames)
    log.info("Setting parkey", srepr(parkey), "in", srepr(rmapping.basename))
    rmapping.header["parkey"] = pktuple
    rmapping.write(new_filename)
    rmapping = rmap.load_mapping(new_filename)
    rmapping = rmap_insert_references_by_matches(new_filename, new_filename, references_headers)
    return rmapping

def get_refactoring_header(rmapping, refname, required_keywords):
    """Create a composite header which is derived from the file contents overidden by any values
    as they appear in the rmap.
    """
    rmapping = rmap.asmapping(rmapping)
    # A fallback source of information is the reference file headers
    header = rmapping.get_refactor_header(
        config.locate_file(refname, rmapping.observatory),
        extra_keys=("META.OBSERVATION.DATE", "META.OBSERVATION.TIME", "DATE-OBS","TIME-OBS") + required_keywords)
    # The primary source of information is the original rmap and the matching values defined there
    headers2 = matches.find_match_paths_as_dict(rmapping.filename, refname)
    # Combine the two,  using the rmap values to override anything duplicated in the reffile header
    assert len(headers2) == 1, "Can't refactor file with more than one match: " + srepr(refname)
    header.update(headers2[0])
    return header

# ============================================================================

def set_rmap_substitution(rmapping, new_filename, parameter_name, old_text, new_text, *args, **keys):
    log.info("Adding substitution for", srepr(parameter_name),
             "from", srepr(old_text), "to", srepr(new_text), "in", srepr(rmapping.basename))
    new_mapping = rmapping.copy()
    if "substitutions" not in new_mapping.header:
        new_mapping.header["substitutions"] = {}
    new_mapping.header["substitutions"][parameter_name] = { old_text : new_text }
    new_mapping.write(new_filename)

# ============================================================================

def cat_rmap(rmapping, new_filename, header_key, *args, **keys):
    """Cat/print rmapping's source text or the value of `header_key` in the rmap header."""
    if header_key is not None:
        log.info("In", srepr(rmapping.basename), "parameter", srepr(header_key), "=", srepr(rmapping.header[header_key]))
    else:
        log.info("-"*80)
        log.info("Rmap", srepr(rmapping.basename), "is:")
        log.info("-"*80)
        log.write(str(rmapping))

# ============================================================================

def add_rmap_useafter(rmapping, new_filename, *args, **keys):
    """Restructure the rmap in Match --> UseAfter form using JWST naming conventions."""
    set_rmap_header(rmapping, new_filename,
                    "classes", repr(rmapping.obs_package.DEFAULT_SELECTORS), *args, **keys)
    parkey = rmapping.obs_package.USEAFTER_KEYWORDS
    if parkey != rmapping.parkey[-1]:
        parkey = rmapping.parkey + (parkey,)
    else:
        parkey = rmapping.parkey
    set_rmap_parkey(rmapping, new_filename, repr(parkey), *args, **keys)
    fix_rmap_undefined_useafter(rmapping, new_filename, *args, **keys)

def fix_rmap_undefined_useafter(rmapping, new_filename, *args, **keys):
    """Change undefined USEAFTER dates to condig.FAKE_USEAFTER_VALUE"""
    rmapping = rmap.ReferenceMapping.from_file(new_filename)
    if "UNDEFINED UNDEFINED" in str(rmapping):
        rmap.replace_rmap_text(rmapping, new_filename, "UNDEFINED UNDEFINED",
                          "1900-01-01 00:00:00", *args, **keys)

# ============================================================================

def insert_rmap_references(rmapping, new_filename, *args, **keys):
    """Insert the appropriate references from the `categorized` references
    dictionary into `rmapping` writing the result to `new_filename`.
    """
    categorized = keys.pop("categorized")
    references = categorized[(rmapping.instrument, rmapping.filekind)]
    rmap_insert_references(rmapping.filename, new_filename, references)

# ============================================================================

def diff_rmap(rmapping, new_filename, *args, **keys):
    """Difference `rmapping` against refactored rmap `new_filename`."""
    script = diff.DiffScript("crds.diff {0} {1} --brief --mapping-text-diffs "
                             "--check-diffs --recurse-added-deleted".format(rmapping.filename, new_filename),
                             reset_log=False, print_status=False)
    script()

def certify_rmap(rmapping, new_filename, source_context=None, *args, **keys):
    """Certify `new_filename` to verify refactored rmap is valid."""
    script = certify.CertifyScript("crds.certify {0} --comparison-context {1}".format(new_filename, rmapping.filename),
                                   reset_log=False, print_status=False)
    script()

# ============================================================================

def apply_rmap_fixers(rmapping, new_filename, fixers, *args, **keys):
    """Apply the text replacements defined in list of colon separated
    old:new `fixers` list to `rmapping` writing results to `new_filename`.
    """
    keys = dict(keys)
    keys.pop("old_text", None)
    keys.pop("new_text", None)
    for fixer in fixers:
        old_text, new_text = fixer.split(":")
        rmap.replace_rmap_text(rmapping, new_filename, old_text, new_text, *args, **keys)
        rmapping = rmap.load_mapping(new_filename)

# ============================================================================

class RefactorScript2(cmdline.Script):
    """Command line script for modifying .rmap files."""

    description = """
    Modifies a reference mapping as indicated by the specified command.
    """

    epilog = """

    The insert_reference and delete_reference commands add references to rmaps
    in a manner similar to that done on the website using Batch Submit References.

    1. Insert reference files A B C D... into rmap X creating rmap Y.

    crds refactor2 insert_reference --old-rmap X --new-rmap Y  --references A B C D...

    2. Delete references A B C D ... from rmap X creating rmap Y.

    crds refactor2 delete_reference --old-rmap Y --new-rmap Y  --references A B C D...

    The set_header, del_header, del_parameter, set_parkey, and replace commands operate on
    the set of rmaps found under a source context which correspond to the specified instruments and types.

    3. Set header key K to value V in rmap X creating rmap Y.
    source context C (e.g. jwst-edit) for types A B C... of instruments X Y Z...

    crds refactor2 set_header --header-key K --header-value V

    4. Delete header key K (e.g. description) in all the rmaps found under
    source context C (e.g. jwst-edit) for types A B C... of instruments X Y Z...

    crds refactor2 del_header --header-key K \\
           --source-context C --instruments X Y Z ... --types A B C ...

    5. Remove single parameter name N (e.g. META.SUBARRAY.NAME) in all the rmaps found under
    source context C (e.g. jwst-edit) for types A B C... of instruments X Y Z...
    This incudes modifying both parkey and all corresponding selector matching patterns.

    crds refactor2 del_parameter --parameter-name N \\
           --source-context C --instruments X Y Z ... --types A B C ...

    6. Set complete parkey in all rmaps found under
    source context C (e.g. jwst-edit) for types A B C... of instruments X Y Z...
    This removes all the existing references and re-inserts them under the new parkey approach.

    crds refactor2 set_parkey --parkey "(('META.INSTRUMENT.DETECTOR','META.SUBARRAY.NAME',))" \\
            --source-context jwst-edit --instruments X Y Z ... --types A B C ... \\
            --fixers FGS1:GUIDER1 FGS2:GUIDER2 ANY:GENERIC FULL:GENERIC

    7. Replace old text P1 with new text P2 in all rmaps found under
    source context C (e.g. jwst-edit) for types A B C... of instruments X Y Z...
    This is a simple text substitution.

    crds refactor2 replace_text --old-text P1  --new-text P2 \\
            --source-context jwst-edit --instruments X Y Z ... --types A B C ...

    8. Add an unconditional load-time parameter value substitution to the rmap header.  For this example,
    the given substitution is added to the rmap "substitutions" header dictionary,  which has the effect
    that match tuples in the rmap are transparently altered at load-time, and SUBARRAY values of GENERIC
    are re-interpreted as N/A for the purposes of matching.

    crds refactor2 set_substitution --parameter-name META.SUBARRAY.NAME  --old-text GENERIC  --new-text N/A \\
            --source-context jwst-edit --instruments X Y Z ... --types A B C ...

    IOW,  this command adds/updates something similar to the following to the specified rmaps:

      'substitutions' : {
          'META.SUBARRAY.NAME' : { 'GENERIC' : 'N/A' },
      }

    9. All of the above commands which elaborate rmaps to refactor based on --source-context can also be
    driven by a direct specification of .rmap names using --rmaps:

    crds refactor2 set_substitution --parameter-name META.SUBARRAY.NAME  --old-text GENERIC  --new-text N/A \\
            --rmaps jwst_miri_dark_0007.rmap

    10. Add the nested UseAfter selector to early JWST rmaps based on Match-only.

    crds refactor2 add_useafter --rmaps jwst_miri_dark_0007.rmap

    11. Mass reference insertion for rehearsing large file deliveries based on --source-context

    It's possible to insert a list of --references into rmaps determined by each reference instrument and type
    and the --source-context.  This mode is active for insert_reference if --old-rmap is not supplied.  If
    the --source-context does not already contain a corresponding rmap for some reference,  CRDS attempts
    to fetch an empty-rmap-spec from the appropriate installed specs directory to use as the baseline rmap.

    This mode allows testing a heterogeneous group of reference types without having to individually specify
    the baseline rmap for each type.  By falling back to the type declaration specs as baseline rmaps, it
    supports reference insertion into rmaps not yet officially added to the --source-context.

    crds refactor2 insert_reference --references ./*.asdf --verbose --certify --diff

    ---------------------------------------------------------------------------------------------------------

    NOTE: In general,  when refactoring,  there are a number of CRDS and JWST calibration code data model
    environment variables which control allowing or passing bad values.   As of now:

    setenv CRDS_ALLOW_BAD_PARKEY_VALUES 1  # leave bad parkey values as-is rather than failing
    setenv CRDS_ALLOW_BAD_USEAFTER      1  # define bad or missing USEAFTER as 1900-01-01T00:00:00
    setenv CRDS_ALLOW_SCHEMA_VIOLATIONS 1  # treat data model schema violations as warnings not errors
    setenv PASS_INVALID_VALUES          1  # ask cal-code data model not to omit invalid values

    See --best-effort for a short cut switch to turn all of the above on from the command line.
    """

    def add_args(self):
        self.add_argument("command", choices=("insert_reference", "delete_reference", "set_header", "set_substitution",
                                              "del_header", "del_parameter", "set_parkey", "replace_text", "cat",
                                              "add_useafter", "diff_rmaps", "certify_rmaps"),
                          help="Name of refactoring command to perform.")
        self.add_argument('--old-rmap', type=cmdline.mapping_spec, default=None,
                          help="Reference mapping to modify by inserting references.")
        self.add_argument('--new-rmap', type=cmdline.mapping_spec, default=None,
                          help="Name of modified reference mapping output file.")
        self.add_argument('--references', type=str, nargs="*",
                          help="Reference files, to insert into (or delete from) `old_rmap` to produce `new_rmap`.")
        self.add_argument('--source-context', type=str, default=None,
                          help="Source context from which to retrieve affected mappings.")
        self.add_argument('--instruments', type=str, nargs="*", help="Instruments to which to apply this operation.")
        self.add_argument('--types', type=str, nargs="*", help="Reference types to which to apply this operation.")
        self.add_argument('--rmaps',  type=str, nargs="*", help="Explicitly specify list of rmaps for refactoring.")
        self.add_argument('--header-key', type=str, default=None, help="Header keyword for header commands.")
        self.add_argument('--header-value', type=str, default=None, help="Value for header commands.")
        self.add_argument('--parkey', type=str, default=None, help="New parkey value for set_parkey.")
        self.add_argument('--parameter-name', type=str, default=None, help="Name of matching parameter to operate on.")
        self.add_argument('--old-text', type=str, default=None, help="Source text for replace_text command.")
        self.add_argument('--new-text', type=str, default=None, help="Replacement text for replace_text command.")
        self.add_argument('--inplace', action="store_true",
                          help="Rewrite files directly in cache replacing original copies.")
        self.add_argument('--fixers', type=str, nargs="*", default=["FGS1:GUIDER1","FGS2:GUIDER2"],
                          help="Simple colon separated global replacements of form old:new ... applied before refactoring.")
        self.add_argument("--sync-files", dest="sync_files", action="store_true",
            help="Fetch any missing files needed for the requested refactoring from the CRDS server.")
        self.add_argument("--diff-rmaps", action="store_true",
                          help="After refactoring, crds.diff the refactored version against the original.")
        self.add_argument("--certify-rmaps", action="store_true",
                          help="After refactoring run crds.certify on refactored rmaps.")
        self.add_argument("--best-effort", action="store_true",
                          help="Shorthand for switching on error bypass env settings.  Generally pass bad values for manual correction later.")

    def main(self):

        if self.args.best_effort:
            config.PASS_INVALID_VALUES.set(True)           # JWST SSB cal code data model
            config.ALLOW_BAD_USEAFTER.set(True)            # Don't fail for bad USEAFTER values
            config.ALLOW_SCHEMA_VIOLATIONS.set(True)       # Don't fail for data model bad value errors
            config.ALLOW_BAD_PARKEY_VALUES.set(True)       # Don't fail for values which don't pass DM + .tpn checking

        if self.args.rmaps:   # clean up dead lines from file lists
            self.args.rmaps = [ self.resolve_context(mapping) for mapping in self.args.rmaps if mapping.strip() ]

        if self.args.references:
            self.args.references = [self.locate_file(reference) for reference in self.args.references]

        with log.error_on_exception("Refactoring operation FAILED"):
            if self.args.command == "insert_reference":
                if self.args.old_rmap:
                    old_rmap, new_rmap = self.resolve_context(self.args.old_rmap), self.resolve_context(self.args.new_rmap)
                    rmap_insert_references(old_rmap, new_rmap, self.args.references)
                else:
                    self.insert_references()  # figure it all out relative to --source-context
            elif self.args.command == "delete_reference":
                old_rmap, new_rmap = self.resolve_context(self.args.old_rmap), self.resolve_context(self.args.new_rmap)
                rmap_delete_references(old_rmap, new_rmap, self.args.references)
            elif self.args.command == "del_header":
                self.del_header_key()
            elif self.args.command == "set_header":
                self.set_header_key()
            elif self.args.command == "del_parameter":
                self.del_parameter()
            elif self.args.command == "set_parkey":
                self.set_parkey()
            elif self.args.command == "replace_text":
                self.replace_text()
            elif self.args.command == "set_substitution":
                self.set_substitution()
            elif self.args.command == "cat":
                self.cat()
            elif self.args.command == "add_useafter":
                self.add_useafter()
            elif self.args.command == "diff_rmaps":
                self.diff_rmaps()
            elif self.args.command == "certify_rmaps":
                self.certify_rmaps()
            else:
                raise ValueError("Unknown refactoring command: " + repr(self.args.command))

        log.standard_status()
        return log.errors()

    def rmap_apply(self, func, *args, **keys):
        """Apply `func()` to *args and **keys,  adding the pmap, imap, and rmap values
        associated with the elaboration of args.source_context, args.instruments, args.types.
        """
        keywords = dict(keys)
        self._setup_source_context()
        if self.args.rmaps:
            for rmap_name in self.args.rmaps:
                with log.error_on_exception("Failed processing rmap", srepr(rmap_name)):
                    log.info("="*20, "Refactoring rmap", srepr(rmap_name), "="*20)
                    rmapping = rmap.load_mapping(rmap_name)
                    new_filename = self._process_rmap(func, rmapping=rmapping, **keywords)
                    self._diff_and_certify(rmapping=rmapping, new_filename=new_filename,
                                           source_context=self.source_context, **keywords)
        else:
            pmapping = rmap.load_mapping(self.source_context)
            instruments = pmapping.selections.keys() if "all" in self.args.instruments else self.args.instruments
            for instr in instruments:
                with log.augment_exception("Failed loading imap for", repr(instr), "from",
                                            repr(self.source_context)):
                    imapping = pmapping.get_imap(instr)
                types = imapping.selections.keys() if "all" in self.args.types else self.args.types
                for filekind in types:
                    with log.error_on_exception("Failed processing rmap for", repr(filekind)):
                        #, "from",
                        # repr(imapping.basename), "of", repr(self.source_context)):
                        try:
                            rmapping = imapping.get_rmap(filekind).copy()
                        except crds.exceptions.IrrelevantReferenceTypeError as exc:
                            log.info("Skipping type", srepr(filekind), "as N/A")
                            continue
                        log.info("="*20, "Refactoring rmap", srepr(rmapping.basename), "="*20)
                        new_filename = self._process_rmap(func, rmapping=rmapping, **keywords)
                        self._diff_and_certify(rmapping=rmapping, source_context=self.source_context,
                                               new_filename=new_filename, **keywords)

    def _diff_and_certify(self, *args, **keys):
        """Apply the diff and/or certify scripts if requested on the command line."""
        if self.args.diff_rmaps:
            diff_rmap(*args, **keys)
        if self.args.certify_rmaps:
            certify_rmap(*args, **keys)

    def _process_rmap(self, func, rmapping, *args, **keys):
        """Execute `func` on a single `rmapping` passing along *args and **keys"""
        keywords = dict(keys)
        rmapping_org = rmapping
        new_filename  = rmapping.filename if self.args.inplace else os.path.join(".", rmapping.basename)
        if os.path.exists(new_filename):
            log.info("Continuing refactoring from local copy", srepr(new_filename))
            rmapping = rmap.load_mapping(new_filename)
        keywords.update(locals())
        fixers = self.args.fixers
        if fixers:
            rmapping = rmap.load_mapping(rmapping.filename)
            keywords.update(locals())
            apply_rmap_fixers(*args, **keywords)
        func(*args, **keywords)
        return new_filename

    def _setup_source_context(self):
        """Default the --source-context if necessary and then translate any symbolic name to a literal .pmap
        name.  e.g.  jwst-edit -->  jwst_0109.pmap.   Then optionally sync the files to a local cache.
        """
        if self.args.source_context is None:
            self.source_context = self.observatory + "-edit"
            log.info("Defaulting --source-context to", srepr(self.source_context))
        else:
            self.source_context = self.args.source_context
        self.source_context = self.resolve_context(self.source_context)
        if self.args.sync_files:
            errs = sync.SyncScript("crds.sync --contexts {}".format(self.source_context))()
            assert not errs, "Errors occurred while syncing all rules to CRDS cache."

    def del_header_key(self):
        """Set args.header_key to string args.header_value in all rmaps elaborated under
        args.source_context, args.instruments, and args.types.
        """
        self.rmap_apply(del_rmap_header, key=self.args.header_key)

    def set_header_key(self):
        """Set args.header_key to string args.header_value in all rmaps elaborated under
        args.source_context, args.instruments, and args.types.
        """
        self.rmap_apply(set_rmap_header, header_key=self.args.header_key, header_value=self.args.header_value)

    def del_parameter(self):
        """Delete args.parameter_name from the parkey of each rmapping in the rmap
        elaboration of the command line parameters.   Somewhat redundant with set_parkey
        which replaces the entire header item,  but doesn't require viable reference
        parameters to support reference file re-insertion.
        """
        self.rmap_apply(del_rmap_parameter, parameter=self.args.parameter_name)

    def set_parkey(self):
        """Change the parkey in rmaps elaborated from command line and automatically
        delete and re-insert each reference to construct appropriate matching tuples.
        """
        self.rmap_apply(set_rmap_parkey, parkey=self.args.parkey)

    def replace_text(self):
        """Do simple text substitution in elaborated rmaps replacing `args.old_text` with `args.new_text`."""
        self.rmap_apply(rmap.replace_rmap_text, old_text=self.args.old_text,  new_text=self.args.new_text)

    def set_substitution(self):
        """Do simple text substitution in elaborated rmaps replacing `args.old_text` with `args.new_text`."""
        self.rmap_apply(set_rmap_substitution, parameter_name=self.args.parameter_name, old_text=self.args.old_text,  new_text=self.args.new_text)

    def cat(self):
        """Either cat the text of the elaborated rmaps if no --parameter-name is specified,  or dump the specified
        --parameter-name from the rmap headers.
        """
        self.rmap_apply(cat_rmap, header_key=self.args.header_key)

    def add_useafter(self):
        """Restructure rmaps to Match -> UseAfter form."""
        self.rmap_apply(add_rmap_useafter)

    def diff_rmaps(self):
        """Apply crds.diff to the refactored rmaps."""
        self.rmap_apply(diff_rmap)

    def certify_rmaps(self):
        """Apply crds.certify to the refactored rmaps."""
        self.rmap_apply(certify_rmap)

    def insert_references(self):
        """Insert files specified by --references into the appropriate rmaps identified by --source-context."""
        self._setup_source_context()
        categorized = self.categorize_files(self.args.references)
        pmap = crds.get_pickled_mapping(self.source_context)  # reviewed
        self.args.rmaps = []
        for (instrument, filekind) in categorized:
            try:
                self.args.rmaps.append(pmap.get_imap(instrument).get_rmap(filekind).filename)
            except crexc.CrdsError:
                log.info("Existing rmap for", (instrument, filekind), "not found.  Trying empty spec.")
                spec_file = os.path.join(
                    os.path.dirname(self.obs_pkg.__file__), "specs", instrument + "_" + filekind + ".rmap")
                rmapping = rmap.asmapping(spec_file)
                log.info("Loaded spec file from", repr(spec_file))
                self.args.rmaps.append(spec_file)
        self.rmap_apply(insert_rmap_references, categorized=categorized)

if __name__ == "__main__":
    sys.exit(RefactorScript2()())



"""This module encapsulates code related to determining the datasets which are affected
by table changes.   The nature of the optimization is that,  viewed as a file change,  CRDS
would think a new table "affects everything" and be forced to recommend reprocessing all datasets.
The code in this module is tasked with determining if the rows selected from two versions of
a table by particular dataset parameters are actually different.

If the rows are not different, then effectifvely the new version of the table should not cause
a dataset to be processed.

If the rows are different,  then the dataset should be reprocessed.
"""
from crds.core import rmap, log
from crds.io import tables
from crds.client import api

# ===================================================================

def is_reprocessing_required(dataset,  dataset_parameters, old_context, new_context, update):
    """This is the top level interface to crds.bestrefs running in "Affected Datasets" mode.

    It determines if reprocessing `dataset` with parameters `dataset_parameters` should be performed as
    a consequence of switching from `old_reference` to `new_reference`.  old_reference is assigned to dataset
    by old_context,  and new_reference is assigned to dataset by new_context.

    Parameters
    ----------
    dataset:
             id of dataset being reprocessed,  <assoc>:<member> or <unassoc>:<unassoc> format

    dataset_parameters:
                        { parameter : value, ...} for all matching parameters and row selection parameters

                        XXX row selection parameters not used in file selection may not be present until
                        XXX explicitly added to the CRDS interface to the DADSOPS parameter database...
                        XXX and possibly even to DADSOPS itself. Normally the row selections have only been
                        XXX done with direct access to dataset .fits files.

    old_context: loaded pmap or name of old context,  possibly for metadata or None

    new_context: loaded pmap or name of new context,  possibly for metadata

    update: Update object

    Returns
    -------
    True        IFF reprocessing should be done as a consequence of the table change.
    """

    log.verbose('is_reprocessing_required: Called with:\n',
                dataset, '\n',
                dataset_parameters, '\n',
                old_context, '\n',
                new_context, '\n',
                update,
                verbosity=100)

    # no old_context means "single context" mode,  always reprocess.
    if old_context is None:
        return True

    # NOTE: non-tables are treated in DeepLook as filekinds which aren't (or maybe someday are) handled,
    # hence reprocessed for now.

    # Reprocess for non-file special values.  Other code will decide what to do with the updates,
    # the point here is that table comparison isn't possible so filtering shouldn't be done.
    old_ref = update.old_reference.lower()
    new_ref = update.new_reference.lower()
    incomparable = ('n/a', 'undefined', 'not found')
    if old_ref.startswith(incomparable) or new_ref.startswith(incomparable):
        return True

    # mostly debug wrappers here,  allows simple string parameters to work and resolves cache paths.
    old_context = rmap.asmapping(old_context, cached=True)
    new_context = rmap.asmapping(new_context, cached=True)
    old_reference = old_context.locate_file(old_ref)
    new_reference = new_context.locate_file(new_ref)

    # Log that deep examination is occuring.
    log.verbose('Deep Reference examination between {} and {} initiated.'.format(old_reference, new_reference),
                verbosity=25)

    with log.error_on_exception("Failed fetching comparison reference tables:", repr([old_ref, new_ref])):
        api.dump_files(new_context.name, [old_ref, new_ref])

    # See if deep checking into the reference is possible.
    try:
        deep_look = DeepLook.from_filekind(update.instrument, update.filekind)

        dataset_id = dataset.split(':')[0]

        # **DEBUG**
        # ** Since we are not getting full headers, if this is a test
        # ** dataset, replace the headers.
        #log.verbose_warning('Forcing use of LBYX01010, regardless...', verbosity=25)
        #dataset_id = 'LBYX01010'           #***DEBUG: force headers regardless of actua data

        if dataset_id in deep_look.stub_input:
            log.verbose_warning('Substituting header for dataset "{}"'.format(dataset))
            dataset_parameters = deep_look.stub_input[dataset_id]['headers']
            log.verbose_warning('headers = ', dataset_parameters, verbosity=25)

        log.verbose(deep_look.preamble, 'Dataset headers = {}'.format(dataset_parameters), verbosity=75)
        log.verbose(deep_look.preamble, 'Comparing references {} and {}.'.format(old_reference, new_reference), verbosity=75)
        deep_look.are_different(dataset_parameters, old_reference, new_reference)

        log.verbose(deep_look.preamble, 'Reprocessing is {}required.'.format('' if deep_look.is_different else 'not '), verbosity=25)
        log.verbose(deep_look.preamble, deep_look.message, verbosity=25)
        return deep_look.is_different

    except DeepLookError as error:

        # Could not determine difference, therefore presume so.
        log.verbose_warning('Deep examination error: {}'.format(error.message), verbosity=25)
        log.verbose_warning('Deep examination failed, presuming reprocessing.', verbosity=25)
        return True


###########
# Utilities
###########
def str_to_number(val, strip=True):
    """Map string `input` to the simplest numerical type capable of parsing it.  If `input`
    will not parse for any type,  return it as-is,  optionally stripping whitespace.
    """

    types = [int, float, complex]

    result = None
    for typ in types:
        try:
            result = typ(val)
            break
        except Exception:
            continue

    if result is None:
        result = val.strip() if strip else val

    return result

def mode_select(table, constraints):
    """Return rows that match the constraints

    Parameters
    ----------
    table: simple table
           Table to examine

    constraints: {field: (value, cmpfn, **kargs}
                 For each field, compare the given value using the
                 the specified comparison function. The cmpfn looks like
                     bool = cmpfn(row[field], value, **kargs)

    Returns
    -------
    The next row that matches.
    """
    for row in table.rows:
        selected = True
        for field in constraints:
            field_index = table.colnames.index(field.upper())
            (value, cmpfn, args) = constraints[field]
            selected = selected & cmpfn(str_to_number(row[field_index]), value, args)

        if selected:
            yield row

def mode_equality(modes_a, modes_b):
    """Check if the modes are equal"""

    # Assume not equal
    equality = False

    # Must be the same length
    if len(modes_a) == len(modes_b):

        # Must have some length
        if len(modes_a) > 0:
            equality = (modes_a == modes_b)

        # Else, nothing compares, so basically equal
        else:
            equality = True

    # That's all folks
    return equality

###################
#
# Comparison functions
#
###################

def cmp_equal(table_value, matching_values, wildcards=()):
    """Value equality

    Parameters
    ----------
    table_value: obj
                 Value from a reference table.

    matching_values: obj or [obj,]
                     What to check against. May be singular or a list

    wildcards: [value,]
               Values that are considered "everything".
               If the table_value is a wildcard, equality is
               presumed.
"""

    # Presume not equal
    is_equal = False

    # Is this a wildcard?
    if table_value in wildcards:
        is_equal = True

    # Otherwise, do a direct match
    else:
        try:
            is_equal = (table_value in matching_values)
        except Exception:
            is_equal = (table_value == matching_values)

    # That's all folks.
    return is_equal



# ##############################
#
# DeepLook
#
# The rules.
#
###############################


class DeepLookError(Exception):
    """Deep Look error base class
    """

    def __init__(self, message):
        super(DeepLookError, self).__init__(message)
        self.message = message

    def __str__(self):
        return repr(self.message)

class DeepLook:
    """Base class to define how reference tables are deep-checked
    for differences between references
    """

    rules = {} # List of classes to use for rules.

    def __init__(self):

        # **DEBUG**
        # Define dummy data.
        # Remove when proper headers are retrievable
        self.stub_input = { # dataset => headers
        }

        # Preamble for log messages
        self.preamble = 'Rule {}:'.format(self.__class__.__name__)

        # Meta Values that may be found in header keywords
        self.metavalues = {}

        # Default mode fields
        self.mode_fields = {}

        # Default way to compare
        self.cmp_equal_parameters = (cmp_equal, {'wildcards': ['ANY']})

        # Basic presumption is that there is a difference unless
        # proven otherwise.
        self.is_different = True
        self.message = 'Comparision not done, presuming references are different.'


    @classmethod
    def from_filekind(cls, instrument, filekind):
        """Create the appropriate object for the type of reference file"""

        name = (instrument + '_' + filekind).lower()
        log.verbose('Instantiating rules for reference type {}.'.format(name), verbosity=25)
        if name in cls.rules:
            return  cls.rules[name]()
        else:
            raise DeepLookError('No rules for instrument {} and reference file kind {}'.format(instrument, filekind))

    def are_different(self, headers, old_reference, new_reference):
        """Do the deep examination of the reference files with-respect-to the given dataset headers

        Affects
        =======
            self.is_different: Sets True or False whether the references are different.
            self.message: Reason for current state of is_different
        """

        # Convert header keys to lowercase for consistency
        headers_low = dict((k.lower(), v) for k, v in headers.items())

        # Start off that the references are different.
        self.is_different = True
        self.message = 'Comparision started but not completed.'

        # Get values for the mode fields
        constraint_values = {}
        for field in self.mode_fields:
            constraint_values[field] = str_to_number(headers_low[field]) if field in headers_low else None
        if None in constraint_values.values():
            self.message = 'Not all mode fields are defined in the dataset.'
            return

        # Modify the constraint values if any "meta" values are
        # present.
        for key in constraint_values:
            if key in self.metavalues:
                if constraint_values[key] in self.metavalues[key]:
                    constraint_values[key] = self.metavalues[key][constraint_values[key]]

        # Read the references
        data_old = tables.tables(old_reference)[0]   # XXXX currently limited to FITS extension 1
        data_new = tables.tables(new_reference)[0]

        # Columns must be the same between tables.
        if sorted(data_old.colnames) != sorted(data_new.colnames):
            self.message = 'Columns are different between references.'
            return

        # Now that values are in hand, produce the full constraint
        # dictionary
        constraints = {}
        for field in self.mode_fields:
            constraints[field] = (constraint_values[field],) + self.mode_fields[field]

        log.verbose(self.preamble, 'Constraints are:\n', constraints, verbosity=75)

        # Reduce the tables to just those rows that match the mode
        # specifications.
        mode_rows_old = [repr(row) for row in mode_select(data_old, constraints)]
        mode_rows_new = [repr(row) for row in mode_select(data_new, constraints)]

        # Sort the rows
        mode_rows_old.sort()
        mode_rows_new.sort()

        log.verbose(self.preamble, 'Old reference matching rows:\n', mode_rows_old, verbosity=75)
        log.verbose(self.preamble, 'New reference matching rows:\n', mode_rows_new, verbosity=75)

        # Check on equality.
        # That's all folks.
        self.is_different = not mode_equality(mode_rows_old, mode_rows_new)

        if self.is_different:
            self.message = 'Selection rules have executed and the selected rows are different.'
        else:
            self.message = 'Selection rules have executed and the selected rows are the same.'


################################
#
# Rule for tables that cannot be examined
#
################################


class DeepLook_Default(DeepLook):
    def are_different(self, headers, old_reference, new_reference):

        self.is_different = True
        self.message = 'Reference type cannot be examined, by definition.'


#############################
#
# Rules for COS
#
#############################


class DeepLook_COS(DeepLook):
    """Generic class for all COS rules
    """

    def __init__(self):
        super(DeepLook_COS, self).__init__()

        # **DEBUG**
        # Define dummy data.
        # Remove when proper headers are retrievable
        self.stub_input = { # dataset => headers
            'xLA7803FKQ': {
                'headers': {
                    'opt_elem': 'G160M',
                    'cenwave':  '1600',
                    'aperture': 'WCA',
                    'segment':  'BOTH',
                },
            },
            'xLBYX01010': {
                'headers': {
                    'opt_elem': 'G140L',
                    'cenwave':  '1280',
                    'aperture': 'PSA',
                    'segment':  'FUVB',
                },
            },
            'xLB4P02050': {
                'headers': {
                    'opt_elem': 'G160M',
                    'cenwave':  '1600',
                    'aperture': 'PSA',
                },
            },
            'xLB4P07010': {
                'headers': {
                    'opt_elem': 'G140L',
                    'cenwave':  '1230',
                    'aperture': 'PSA',
                },
            },
            'xLB6M01030': {
                'headers': {
                    'opt_elem': 'G230L',
                    'cenwave':  '3000',
                    'aperture': 'PSA',
                },
            },
            'xLBK617010': {
                'headers': {
                    'opt_elem': 'G185M',
                    'cenwave':  '1986',
                    'aperture': 'PSA',
                },
            },
        }


class DeepLook_COSSegment(DeepLook_COS):
    """Tables that require SEGMENT only"""

    def __init__(self):
        super(DeepLook_COSSegment, self).__init__()

        self.mode_fields = {
            'segment': self.cmp_equal_parameters,
        }


class DeepLook_COSFullmode(DeepLook_COS):
    def __init__(self):
        super(DeepLook_COSFullmode, self).__init__()

        self.mode_fields = {
            'opt_elem': self.cmp_equal_parameters,
            'cenwave':  self.cmp_equal_parameters,
            'aperture': self.cmp_equal_parameters,
        }


class DeepLook_COSOpt_elem(DeepLook_COS):
    def __init__(self):
        super(DeepLook_COSOpt_elem, self).__init__()

        self.mode_fields = {
            'opt_elem': self.cmp_equal_parameters,
        }


class DeepLook_COSDISPTAB(DeepLook_COS):
    def __init__(self):
        super(DeepLook_COSDISPTAB, self).__init__()

        self.mode_fields = {
            'opt_elem': self.cmp_equal_parameters,
            'cenwave':  self.cmp_equal_parameters,
        }


class DeepLook_COSLAMPTAB(DeepLook_COS):
    def __init__(self):
        super(DeepLook_COSLAMPTAB, self).__init__()

        self.mode_fields = {
            'opt_elem': self.cmp_equal_parameters,
            'cenwave':  self.cmp_equal_parameters,
            'fpoffset': self.cmp_equal_parameters,
        }


class DeepLook_COSTDSTAB(DeepLook_COS):
    def __init__(self):
        super(DeepLook_COSTDSTAB, self).__init__()

        self.mode_fields = {
            'opt_elem': self.cmp_equal_parameters,
            'aperture': self.cmp_equal_parameters,
        }


class DeepLook_COSFullSegment(DeepLook_COS):
    def __init__(self):
        super(DeepLook_COSFullSegment, self).__init__()

        self.mode_fields = {
            'opt_elem': self.cmp_equal_parameters,
            'cenwave':  self.cmp_equal_parameters,
            'aperture': self.cmp_equal_parameters,
            'segment': self.cmp_equal_parameters,
        }



#############################
#
# Rules for STIS
#
#############################


class DeepLook_STIS(DeepLook):
    """Generic class for all STIS rules
    """

    def __init__(self):
        super(DeepLook_STIS, self).__init__()


class DeepLook_STISopt_elem(DeepLook_STIS):
    def __init__(self):
        super(DeepLook_STISopt_elem, self).__init__()

        self.mode_fields = {
            'opt_elem': self.cmp_equal_parameters,
        }

class DeepLook_STISaperture(DeepLook_STIS):
    def __init__(self):
        super(DeepLook_STISaperture, self).__init__()

        self.mode_fields = {
            'aperture': self.cmp_equal_parameters,
        }

class DeepLook_STIScenwave(DeepLook_STIS):
    def __init__(self):
        super(DeepLook_STIScenwave, self).__init__()

        self.mode_fields = {
            'opt_elem': self.cmp_equal_parameters,
            'cenwave': self.cmp_equal_parameters,
        }

class DeepLook_STISfullmode(DeepLook_STIS):
    def __init__(self):
        super(DeepLook_STISfullmode, self).__init__()

        self.mode_fields = {
            'opt_elem': self.cmp_equal_parameters,
            'cenwave':  self.cmp_equal_parameters,
            'aperture': self.cmp_equal_parameters,
        }

class DeepLook_STISCCDTAB(DeepLook_STIS):
    def __init__(self):
        super(DeepLook_STISCCDTAB, self).__init__()

        self.mode_fields = {
            'ccdamp': self.cmp_equal_parameters,
            'ccdgain': self.cmp_equal_parameters,
            'ccdoffst': self.cmp_equal_parameters,
            'binaxis1': self.cmp_equal_parameters,
            'binaxis2': self.cmp_equal_parameters,
        }

class DeepLook_STISLAMPTAB(DeepLook_STIS):
    def __init__(self):
        super(DeepLook_STISLAMPTAB, self).__init__()

        self.mode_fields = {
            'opt_elem': self.cmp_equal_parameters,
            'lampset': self.cmp_equal_parameters,
            'sclamp': self.cmp_equal_parameters,
        }

class DeepLook_STISMLINTAB(DeepLook_STIS):
    def __init__(self):
        super(DeepLook_STISMLINTAB, self).__init__()

        self.mode_fields = {
            'detector': self.cmp_equal_parameters,
        }

class DeepLook_STISWCPTAB(DeepLook_STIS):
    def __init__(self):
        super(DeepLook_STISWCPTAB, self).__init__()

        self.mode_fields = {
            'opt_elem': self.cmp_equal_parameters,
            'detector': self.cmp_equal_parameters,
        }


#############################
#
# Rules for ACS
#
#############################


class DeepLook_ACS(DeepLook):
    """Generic class for all STIS rules
    """

    def __init__(self):
        super(DeepLook_ACS, self).__init__()

class DeepLook_ACSCCDpars(DeepLook_ACS):
    def __init__(self):
        super(DeepLook_ACSCCDpars, self).__init__()

        self.mode_fields = {
            'ccdamp':   self.cmp_equal_parameters,
            'ccdchip':  self.cmp_equal_parameters,
            'ccdgain':  self.cmp_equal_parameters,
        }


class DeepLook_ACSDetector(DeepLook_ACS):
    def __init__(self):
        super(DeepLook_ACSDetector, self).__init__()

        self.mode_fields = {
            'detector': self.cmp_equal_parameters,
        }


class DeepLook_ACSCCDTAB(DeepLook_ACSCCDpars):
    def __init__(self):
        super(DeepLook_ACSCCDTAB, self).__init__()

        self.mode_fields.update({
            'binaxis1': self.cmp_equal_parameters,
            'binaxis2': self.cmp_equal_parameters,
        })

#############################
#
# Rules for WFC3
#
#############################


class DeepLook_WFC3(DeepLook):
    """Generic class for all STIS rules
    """

    def __init__(self):
        super(DeepLook_WFC3, self).__init__()


class DeepLook_WFC3BPIXTAB(DeepLook_WFC3):
    def __init__(self):
        super(DeepLook_WFC3BPIXTAB, self).__init__()

        self.mode_fields = {
            'ccdchip':  self.cmp_equal_parameters,
        }


###############
#
# Rule lookup  table
#
################

DeepLook.rules = {
    'cos_badttab':   DeepLook_Default,
    'cos_bpixtab':   DeepLook_COSSegment,
    'cos_brftab':    DeepLook_Default,
    'cos_brsttab':   DeepLook_COSSegment,
    'cos_deadtab':   DeepLook_COSSegment,
    'cos_disptab':   DeepLook_COSDISPTAB,
    'cos_fluxtab':   DeepLook_COSFullmode,
    'cos_gsagtab':   DeepLook_Default,
    'cos_hvtab':     DeepLook_Default,
    'cos_lamptab':   DeepLook_COSLAMPTAB,
    'cos_phatab':    DeepLook_COSOpt_elem,
    'cos_proftab':   DeepLook_COSFullSegment,
    'cos_spwcstab':  DeepLook_COSFullmode,
    'cos_tdstab':    DeepLook_COSTDSTAB,
    'cos_tracetab':  DeepLook_COSFullSegment,
    'cos_twozxtab':  DeepLook_COSFullSegment,
    'cos_walktab':   DeepLook_COSSegment,
    'cos_wcptab':    DeepLook_COSOpt_elem,
    'cos_xtractab':  DeepLook_COSFullmode,

    'stis_apdestab': DeepLook_STISaperture,
    'stis_apertab':  DeepLook_STISaperture,
    'stis_bpixtab':  DeepLook_STISopt_elem,
    'stis_ccdtab':   DeepLook_STISCCDTAB,
    'stis_cdstab':   DeepLook_STISopt_elem,
    'stis_crrejtab': DeepLook_Default,
    'stis_disptab':  DeepLook_STIScenwave,
    'stis_echsctab': DeepLook_STISopt_elem,
    'stis_exstab':   DeepLook_STISopt_elem,
    'stis_gactab':   DeepLook_Default,
    'stis_halotab':  DeepLook_STISopt_elem,
    'stis_idctab':   DeepLook_Default,
    'stis_inangtab': DeepLook_Default,
    'stis_lamptab':  DeepLook_STISLAMPTAB,
    'stis_mlintab':  DeepLook_STISMLINTAB,
    'stis_mofftab':  DeepLook_Default,
    'stis_pctab':    DeepLook_Default,
    'stis_phottab':  DeepLook_STIScenwave,
    'stis_riptab':   DeepLook_STISopt_elem,
    'stis_sdctab':   DeepLook_STISfullmode,
    'stis_sptrctab': DeepLook_STIScenwave,
    'stis_srwtab':   DeepLook_STISopt_elem,
    'stis_tdctab':   DeepLook_Default,
    'stis_tdstab':   DeepLook_STISopt_elem,
    'stis_teltab':   DeepLook_STISopt_elem,
    'stis_wcptab':   DeepLook_STISWCPTAB,
    'stis_xtractab': DeepLook_STISfullmode,

    'acs_atodtab':   DeepLook_ACSCCDpars,
    'acs_bpixtab':   DeepLook_ACSCCDpars,
    'acs_ccdtab':    DeepLook_ACSCCDTAB,
    'acs_crrejtab':  DeepLook_Default,
    'acs_idctab':    DeepLook_Default,
    'acs_imphttab':  DeepLook_Default,
    'acs_mdriztab':  DeepLook_Default,
    'acs_mlintab':   DeepLook_ACSDetector,
    'acs_oscntab':   DeepLook_Default,
    'acs_pctetab':   DeepLook_Default,
    'acs_spottab':   DeepLook_Default,

    'wfc3_atodtab':  DeepLook_ACSCCDpars,
    'wfc3_bpixtab':  DeepLook_WFC3BPIXTAB,
    'wfc3_ccdtab':   DeepLook_ACSCCDTAB,
    'wfc3_crrejtab': DeepLook_Default,
    'wfc3_idctab':   DeepLook_Default,
    'wfc3_imphttab': DeepLook_Default,
    'wfc3_mdriztab': DeepLook_Default,
    'wfc3_oscntab':  DeepLook_Default,
}



"""This module defines functions for modifying rmaps in various ways,  generally
the transformations required to automate rmap maintenance on the CRDS website.
"""
import sys
import os.path

# ============================================================================

from crds.core import exceptions, rmap, log, cmdline
from crds.core.log import srepr
from crds import diff

# ============================================================================

class NoUseAfterError(ValueError):
    "The specified UseAfter datetime didn't exist in the rmap."

class NoMatchTupleError(ValueError):
    "The specified Match tuple didn't exist in the rmap."

# ============================================================================

def set_header_value(old_rmap, new_rmap, key, new_value):
    """Set the value of `key` in `filename` to `new_value` and rewrite the rmap.
    This is potentially lossy since rewriting the rmap may/will lose comments and
    formatting quirks.
    """
    mapping = rmap.load_mapping(old_rmap)
    mapping.header[key] = new_value
    mapping.write(new_rmap)

def del_header_value(old_rmap, new_rmap, key):
    """Set the value of `key` in `filename` to `new_value` and rewrite the rmap.
    This is potentially lossy since rewriting the rmap may/will lose comments and
    formatting quirks.
    """
    mapping = rmap.load_mapping(old_rmap)
    del mapping.header[key]
    mapping.write(new_rmap)

def update_derivation(new_path, old_basename=None):
    """Set the 'derived_from' and 'name' header fields of `new_path`.
    This function works for all Mapping classes:  pmap, imap, and rmap.
    """
    new = rmap.fetch_mapping(new_path)
    if old_basename is None:    # assume new is a copy of old, with old's name in header
        derived_from = new.name
    else:
        derived_from = old_basename
    new.header["derived_from"] = str(derived_from)
    new.header["name"] = str(os.path.basename(new_path))
    new.write(new_path)
    return str(derived_from)

# ============================================================================

def rmap_insert_references(old_rmap, new_rmap, inserted_references):
    """Given the full path of starting rmap `old_rmap`,  modify it by inserting
    or replacing all files in `inserted_references` and write out the result to
    `new_rmap`.    If no actions are performed, don't write out `new_rmap`.

    old_rmap  str     Filepath of source rmap into which `inserted_references` will be inserted
    new_rmap  str     Filepath of updated rmap written out
    inserted_references [ str, ...]    List of reference filepaths to be insterted

    Note that "inserting" a reference file can result in:

    1. adding a new match case,
    2. adding a new USEAFTER case
    3. exactly replacing an existing reference file.

    Other outcomes are also possible for non-standard rmap selector class configurations.

    Additional checking:

    1. Generates an ERROR if any of the inserted reference files have identical
    matching criteria since only one file with those criteria would be added to
    the rmap and the other(s) would be "replaced" by their own insertion set.
    Note: it is valid/common for an inserted reference to replace a reference
    which is already in `old_rmap`.  This ERROR only applies to equalities
    within the inserted_references list.

    2. Generates a WARNING if the matching criteria of any inserted reference
    file is a proper subset of inserted or existing references.  Thes subsets
    will generally lead to the addition of new matching cases.  Since CRDS
    inherited instances of these "subset overlaps" from HST CDBS, this warning
    is only visible with --verbose for HST, they exist.  Since this condition
    is bad both for understanding rmaps and for runtime complexity and
    performance, for JWST the warning is visible without --verbose and will
    also generate a runtime ERROR.  For JWST there is the expectation that an
    offending file submission will either be (a) cancelled and corrected or (b)
    provisionally accepted followed by an immediate manual rmap correction.
    Provisional acceptance gives the option of f keeping the work
    associated with large deliveries where the corrective measure might be to
    manually merge overlapping categories with rmap edits.

    Return None,  `new_rmap` is already the implicit result
    """
    new = old = rmap.fetch_mapping(old_rmap, ignore_checksum=True)
    inserted_cases = {}
    for reference in inserted_references:
        log.info("Inserting", os.path.basename(reference), "into", repr(new.name))
        new = new.insert_reference(reference)
        baseref = os.path.basename(reference)
        with log.warn_on_exception("Failed checking rmap update for", repr(baseref)):
            cases = new.file_matches(baseref)
            for fullcase in cases:
                case = fullcase[1:]
                if case not in inserted_cases:
                    inserted_cases[case] = baseref
                else:
                    log.error("-"*40 + "\nBoth", srepr(baseref),
                              "and", srepr(inserted_cases[case]),
                              "identically match case:\n", log.PP(case), """
Each reference would replace the other in the rmap.
Either reference file matching parameters need correction
or additional matching parameters should be added to the rmap
to enable CRDS to differentiate between the two files.
See the file submission section of the CRDS server user's guide here:
    https://jwst-crds.stsci.edu/static/users_guide/index.html
for more explanation.""")

    new.header["derived_from"] = old.basename
    log.verbose("Writing", repr(new_rmap))
    new.write(new_rmap)

def rmap_delete_references(old_rmap, new_rmap, deleted_references):
    """Given the full path of starting rmap `old_rmap`,  modify it by deleting
    all files in `deleted_references` and write out the result to
    `new_rmap`.    If no actions are performed, don't write out `new_rmap`.

    Return new ReferenceMapping named `new_rmap`
    """
    new = old = rmap.fetch_mapping(old_rmap, ignore_checksum=True)
    for reference in deleted_references:
        log.info("Deleting", repr(reference), "from", repr(new.name))
        new = new.delete(reference)
    new.header["derived_from"] = old.basename
    log.verbose("Writing", repr(new_rmap))
    new.write(new_rmap)
    formatted = new.format()
    for reference in deleted_references:
        reference = os.path.basename(reference)
        assert reference not in formatted, \
            "Rules update failure.  Deleted" + repr(reference) + " still appears in new rmap."
    return new

def rmap_check_modifications(old_rmap, new_rmap, old_ref, new_ref, expected=("add",)):
    """Check the differences between `old_rmap` and `new_rmap` and make sure they're
    limited to the types listed in `expected`.

    expected should be "add" or "replace".

    Returns as_expected,  True IFF all rmap modifications match `expected`.
    """
    diffs = diff.mapping_diffs(old_rmap, new_rmap)
    as_expected = True
    for difference in diffs:
        actual = diff.diff_action(difference)
        if actual in expected:
            pass   # white-list so it will fail when expected is bogus.
        else:
            log.error("Expected one of", repr(expected), "but got", repr(actual),
                      "from change", repr(difference))
            as_expected = False
    with open(old_rmap) as pfile:
        old_count = len([line for line in pfile.readlines() if os.path.basename(old_ref) in line])
    with open(new_rmap) as pfile:
        new_count = len([line for line in pfile.readlines() if os.path.basename(new_ref) in line])
    if "replace" in expected and old_count != new_count:
        log.error("Replacement COUNT DIFFERENCE replacing", repr(old_ref), "with", repr(new_ref), "in", repr(old_rmap),
                  old_count, "vs.", new_count)
        as_expected = False
    return as_expected

# ============================================================================

class RefactorScript(cmdline.Script):
    """Command line script for modifying .rmap files."""

    description = """
    Modifies a reference mapping by adding the specified reference files.
    """

    epilog = """
    """

    locate_file = cmdline.Script.locate_file_outside_cache

    def add_args(self):
        self.add_argument("command", choices=("insert", "delete", "set_header", "del_header"),
            help="Name of refactoring command to perform.")
        self.add_argument('old_rmap', type=cmdline.reference_mapping,
            help="Reference mapping to modify by inserting references.")
        self.add_argument('new_rmap', type=cmdline.reference_mapping,
            help="Name of modified reference mapping output file.")
        self.add_argument('references', type=str, nargs="+",
            help="Reference files to insert into (or delete from) `old_rmap` to produce `new_rmap`.")

    @property
    def old_rmap(self):
        return self.locate_file(self.args.old_rmap)

    @property
    def new_rmap(self):
        return self.locate_file(self.args.new_rmap)

    @property
    def ref_paths(self):
        self.args.files = self.args.references
        return self.files  # standard file location and @-handling for self.args.files

    def main(self):
        with log.error_on_exception("Refactoring operation FAILED"):
            if self.args.command == "insert":
                rmap_insert_references(self.old_rmap, self.new_rmap, self.ref_paths)
            elif self.args.command == "delete":
                rmap_delete_references(self.old_rmap, self.new_rmap, self.ref_paths)
            elif self.args.command == "set_header":
                field, setting = self.args.references[0], " ".join(self.args.references[1:])
                set_header_value(self.old_rmap, self.new_rmap, field, setting)
            elif self.args.command == "del_header":
                field = self.args.references[0]
                del_header_value(self.old_rmap, self.new_rmap, field)
            else:
                raise ValueError("Unknown refactoring command: " + repr(self.args.command))
        log.standard_status()
        return log.errors()

if __name__ == "__main__":
    sys.exit(RefactorScript()())



"""This module is used to generate the SYSTEM CRDSCFG reference file.  The code
is directly edited and the module itself is run to emit a new reference as in:

$ python -m crds.jwst.gen_system_crdscfg  updated_inputs.yaml  >new_reference.yaml

Suggested approach for obtaining updated inputs is downloading the most current
SYSTEM CRDSCFG reference file and updating MANUAL sections.

Current calibration code must be available in order to automatically
regenerates since CRDS will inspect the installed .cfg files to determine Step
configurations and required reference types.
"""
import sys
import re
import fnmatch
import os.path

# ----------------------------------------------------------------------------------------------

import yaml

# ----------------------------------------------------------------------------------------------

from jwst import __version__ as jwst_version
from jwst.stpipe import Pipeline

# ----------------------------------------------------------------------------------------------

import crds
from crds.core import log, exceptions, utils, timestamp, pysh
from crds.core.log import srepr

# ----------------------------------------------------------------------------------------------
VERSION_RE_STR = r"(\d+\.\d+.\d+).*"

CAL_VER = re.match(r"^" + VERSION_RE_STR, jwst_version).group(1)

GENERATION_DATE = timestamp.now("T").split(".")[0]

REFERENCE_DIVIDER = "# vvvvvvvv GENERATED vvvvvvvv"

# --------------------------------------------------------------------------------------

class CrdsCfgGenerator:
    """Used to re-generate SYSTEM CRDSCFG reference files based on modified
    versions of previous reference files.
    """
    def __init__(self, input_yaml):
        self.input_yaml = input_yaml
        self.loaded_cfg = self.get_body()
        self.levels = self.loaded_cfg.levels
        self.exp_types = self.loaded_cfg.exp_types
        self.pipeline_cfgs_to_steps = {}
        self.steps_to_reftypes = {}
        self.generate_pipeline_info()
        self.exptypes_to_cfgs = { exp_type : self.exptype_to_pipelines(exp_type)
                                  for exp_type in self.exp_types }
        self.exptypes_to_reftypes = { exp_type : self.exptype_to_reftypes(exp_type)
                                      for exp_type in self.exp_types }

    def get_body(self):
        """Load the input_yaml as a CRDS Struct and return it."""
        return utils.Struct(yaml.safe_load(self.input_yaml))

    def get_updated_yaml(self):
        """Modify the input_yaml to replace the cal code version and generation date,
        useful for updating an existing reference and running it through this generator.
        """
        input_body = []
        for line in self.input_yaml.splitlines():
            line2 = line.strip()
            if line2.startswith("calibration_software_version:"):
                input_body += ["    calibration_software_version: " + CAL_VER]
            elif line2.startswith("generation_date:"):
                input_body += ["    generation_date: " + GENERATION_DATE]
            elif line2.startswith("crds_version:"):
                input_body += ["    crds_version: " + crds.__version__]
            else:
                input_body += [line]
        return "\n".join(input_body).split(REFERENCE_DIVIDER)[0] + "\n" + REFERENCE_DIVIDER + "\n"

    def generate_pipeline_info(self):
        """Based on the input YAML and JWST cal code,  generate the mappings:
        pipeline_cfgs_to_steps
        steps_to_reftypes
        """
        pysh.sh("rm -rf configs")
        pysh.sh("collect_pipeline_cfgs configs")
        self.pipeline_cfgs_to_steps["skip_2b.cfg"] = []
        for pipeline_cfg in self.loaded_cfg.pipeline_cfgs:
            log.info("Processing", repr(pipeline_cfg))
            cfgdir = "configs" # os.path.dirname(pipepkg.__file__) or ""
            cfgpath = os.path.join(cfgdir, pipeline_cfg)
            p = Pipeline.from_config_file(cfgpath)
            steps_to_reftypes = {}
            for name, stepcfg in p.steps.items():
                if stepcfg.get("skip", True):
                    log.info("Considering", repr(name), "skip")
                else:
                    log.info("Considering", repr(name), "keep")
                    step = p.step_defs[name]  # class
                    if isinstance(step.reference_file_types, list):
                        steps_to_reftypes[name] = step.reference_file_types
                    else:
                        log.warning(repr(name), " is not a Step. Setting reference list to empty.")
                        steps_to_reftypes[name] = []
            self.pipeline_cfgs_to_steps[pipeline_cfg] = sorted(list(steps_to_reftypes.keys()))
            self.steps_to_reftypes.update(steps_to_reftypes)

    def generate_output_yaml(self):
        """Generate the SYSTEM CRDSCFG reference YAML."""
        output_yaml = self.get_updated_yaml()
        output_yaml += yaml.dump({"pipeline_cfgs_to_steps" : self.pipeline_cfgs_to_steps}) + "\n"
        output_yaml += yaml.dump({"steps_to_reftypes" : self.steps_to_reftypes}) + "\n"
        output_yaml += yaml.dump({"exptypes_to_pipelines" : self.exptypes_to_cfgs}) + "\n"
        output_yaml += yaml.dump({"exptypes_to_reftypes" : self.exptypes_to_reftypes}) + "\n"
        return output_yaml

    def __str__(self):
        return self.generate_output_yaml()

    def exptype_to_pipelines(self, exp_type):
        """For a given EXP_TYPE string, return a list of reftypes needed to process that
        EXP_TYPE through the data levels appropriate for that EXP_TYPE.

        Return [reftypes... ]
        """
        pipelines = []
        for level in self.levels:
            pipelines.extend(self.get_level_pipeline(level, exp_type))
        return pipelines

    def exptype_to_reftypes(self, exp_type):
        """Return all reftypes associated with processing all steps of all pipelines for `exp_type`."""
        # with log.error_on_exception("Failed exptype_to_reftypes for", srepr(exp_type)):
        reftypes = []
        for pipeline in self.exptype_to_pipelines(exp_type):
            reftypes.extend(self.get_pipeline_types(pipeline, exp_type))
        reftypes = sorted(list(set(reftypes)))
        return reftypes

    def get_level_pipeline(self, level, exp_type):
        """Interpret the level_pipeline_exptypes data structure relative to
        processing `level` and `exp_type` to determine a pipeline .cfg file.

        Return [ pipeline .cfg ]  or  []
        """
        pipeline_exptypes = self.loaded_cfg.level_pipeline_exptypes[level]
        for mapping in pipeline_exptypes:
            for pipeline, exptypes in mapping.items():
                for exptype_pattern in exptypes:
                    if glob_match(exptype_pattern, exp_type):
                        return [pipeline]
        log.error("Unhandled EXP_TYPE", srepr(exp_type), "for", srepr(level))
        return []

        # raise exceptions.CrdsPipelineCfgDeterminationError("Unhandled EXP_TYPE", srepr(exp_type))

    def get_pipeline_types(self, pipeline, exp_type):
        """Based on a pipeline .cfg filename and an EXP_TYPE,  look up
        the Steps corresponding to the .cfg and extrapolate those to the
        reftypes used by those Steps.   If there are exceptions to the
        reftypes assigned for a particular Step that depend on EXP_TYPE,
        return the revised types for that Step instead.

        Return [reftypes, ...]
        """
        steps = self.pipeline_cfgs_to_steps[pipeline]
        reftypes = []
        for step in steps:
            if step not in self.loaded_cfg.steps_to_reftypes_exceptions:
                reftypes.extend(self.steps_to_reftypes[step])
            else:
                for case in self.loaded_cfg.steps_to_reftypes_exceptions[step]:
                    item = list(case.values())[0]
                    more_reftypes = item["reftypes"][:]
                    exptypes = item["exp_types"][:]
                    found = False
                    for exptype_pattern in exptypes:
                        if glob_match(exptype_pattern, exp_type):
                            log.verbose("Adding exceptional types", more_reftypes,
                                        "for step", srepr(step), "case", srepr(exptype_pattern),
                                        "based on exp_type", srepr(exp_type))
                            found = True
                            reftypes.extend(more_reftypes)
                            break
                    if found:
                        break
                else:
                    raise exceptions.CrdsPipelineTypeDeterminationError("Unhandled EXP_TYPE for exceptional Step", srepr(step))
        return reftypes

# --------------------------------------------------------------------------------------

def glob_match(expr, value):
    """Convert the given glob `expr` to a regex and match it to `value`."""
    re_str = fnmatch.translate(expr)
    return re.match(re_str, value)

# --------------------------------------------------------------------------------------

if __name__ == "__main__":
    with open(sys.argv[1]) as infile:
        print(CrdsCfgGenerator(infile.read()))



import pickle

import crds
from crds.core import log

def main():
    p = crds.get_cached_mapping("hst.pmap")
    s = pickle.dumps(p)
    q = pickle.loads(s)

    p._trace_compare(q)

    log.divider("p == q --> " + repr(p == q))

    log.divider("__getstate__ --> " + repr(p.__getstate__() == q.__getstate__()))

    log.divider("rmap __getstate__ --> " + repr(p.get_imap("acs").get_rmap("biasfile").__getstate__() == q.get_imap("acs").get_rmap("biasfile").__getstate__()))

if __name__ == "__main__":
    main()



"""This module contains doctests and unit tests that exercise the newcontext
module responsible for automatically generating imaps and pmaps given a list
of new rmaps.
"""
import os, os.path
from pprint import pprint as pp

from crds.core import rmap, log, exceptions, pysh
from crds.refactoring import newcontext
from crds import diff
from crds import tests
from crds.tests import test_config

from nose.tools import assert_raises, assert_true

# ==================================================================================

def dt_fake_name():
    """
    Fake names are only used by crds.newcontext when it is run from the command line.

    >>> old_state = test_config.setup()
    >>> os.environ["CRDS_MAPPATH_SINGLE"] = test_config.TEST_DATA

    >>> newcontext.fake_name("data/hst.pmap")
    './hst_0003.pmap'

    >>> newcontext.fake_name("data/hst_cos_0001.imap")
    './hst_cos_0001.imap'

    >>> newcontext.fake_name("data/hst_cos_deadtab_9999.rmap")
    './hst_cos_deadtab_10000.rmap'

    >>> test_config.cleanup(old_state)
    """

def dt_new_context():
    """>>> old_state = test_config.setup()
    >>> os.environ["CRDS_MAPPATH_SINGLE"] = test_config.TEST_DATA

    >>> newcontext.NewContextScript("newcontext.py hst.pmap data/hst_cos_deadtab_9999.rmap data/hst_acs_imphttab_9999.rmap")()
    CRDS - INFO -  Replaced 'hst_acs_imphttab.rmap' with 'data/hst_acs_imphttab_9999.rmap' for 'imphttab' in 'data/hst_acs.imap' producing './hst_acs_10000.imap'
    CRDS - INFO -  Replaced 'hst_cos_deadtab.rmap' with 'data/hst_cos_deadtab_9999.rmap' for 'deadtab' in 'data/hst_cos.imap' producing './hst_cos_0001.imap'
    CRDS - INFO -  Replaced 'data/hst_acs.imap' with './hst_acs_10000.imap' for 'ACS' in 'hst.pmap' producing './hst_0003.pmap'
    CRDS - INFO -  Replaced 'data/hst_cos.imap' with './hst_cos_0001.imap' for 'COS' in 'hst.pmap' producing './hst_0003.pmap'
    CRDS - INFO -  Adjusting name 'hst_acs_10000.imap' derived_from 'hst_acs.imap' in './hst_acs_10000.imap'
    CRDS - INFO -  Adjusting name 'hst_cos_0001.imap' derived_from 'hst_cos.imap' in './hst_cos_0001.imap'
    CRDS - INFO -  Adjusting name 'hst_0003.pmap' derived_from 'hst.pmap' in './hst_0003.pmap'
    0

    XXXX Changes to newcontext intended to make it less error prone in actual
    usage resulted in a generated context which intentionally omits paths from
    all mapping file references.  This makes them harder to load for this test
    so dropping this.

    >> pp([difference[-1] for difference in diff.mapping_diffs("data/hst.pmap", "./hst_0003.pmap")])
    ["replaced 'w3m17170j_imp.fits' with 'xb61855jj_imp.fits'",
     "replaced 's7g1700gl_dead.fits' with 's7g1700gm_dead.fits'",
     "replaced 'hst_acs_imphttab.rmap' with 'data/hst_acs_imphttab_9999.rmap'",
     "replaced 'hst_cos_deadtab.rmap' with 'data/hst_cos_deadtab_9999.rmap'",
     "replaced 'data/hst_acs.imap' with './hst_acs_10000.imap'",
     "replaced 'data/hst_cos.imap' with './hst_cos_0001.imap'"]

    >>> pysh.sh("rm \./*\.[pir]map")
    0

    >>> test_config.cleanup(old_state)

    """

class TestNewContext(test_config.CRDSTestCase):
    '''
    def test_get_imap_except(self):
        r = rmap.get_cached_mapping("hst.pmap")
        with self.assertRaises(exceptions.CrdsUnknownInstrumentError):
            r.get_imap("foo")
    '''

# ==================================================================================


def main():
    """Run module tests,  for now just doctests only."""
    from crds.tests import test_newcontext, tstmod
    import unittest

    suite = unittest.TestLoader().loadTestsFromTestCase(TestNewContext)
    unittest.TextTestRunner().run(suite)

    old_state = test_config.setup()
    result = tstmod(test_newcontext)
    test_config.cleanup(old_state)

    return result

if __name__ == "__main__":
    print(main())



"""This module is used to profile getrecommendations() and bestreds.BestrefsScript()."""
from crds.core import utils, log, heavy_client
from crds import bestrefs
from crds.tests.test_config import run_and_profile

HST_HEADER = {
    'INSTRUME' : 'ACS',
    'DETECTOR': 'SBC',
    'CCDAMP': 'N/A',
    'FILTER1' : 'PR110L',
    'FILTER2' : 'N/A',
    'OBSTYPE': 'SPECTROSCOPIC',
    'FW1OFFST' : 'N/A',
    'FW2OFFST': 'N/A',
    'FWOFFST': 'N/A',
    'DATE-OBS': '1991-01-01',
    'TIME-OBS': '00:00:00'
    }

JWST_HEADER = {
    "meta.instrument.name": "NIRISS",
    "meta.observation.date": "2012-07-25",
    "meta.observation.time": "00:00:00",
    "meta.instrument.detector" : "NIS",
    "meta.instrument.filter" : "F380M",
    "meta.subarray.name" : "FULL",
    "meta.exposture.readpatt" : "FAST",
    "meta.exposure.type" : "NIS_IMAGE",
    }

if __name__ == "__main__":
    run_and_profile("JWST getrecommendations()", 'heavy_client.getrecommendations(JWST_HEADER, observatory="jwst", ignore_cache=False)', globals())
    run_and_profile("HST getrecommendations()", 'heavy_client.getrecommendations(HST_HEADER, observatory="hst", ignore_cache=False)', globals())
    run_and_profile("HST bestrefs file", "bestrefs.BestrefsScript('crds.bestrefs --files data/j8bt09jcq_raw.fits --log-time --stats')()", globals())

# run_and_profile("JWST bestrefs file", "bestrefs.BestrefsScript('crds.bestrefs --files data/ --log-time --stats')()")



"""This module provides functions that interface with the JWST calibration code to determine
things like "reference types used by a pipeline."

>>> header_to_reftypes(test_header("0.7.0", "FGS_DARK"))
['ipc', 'linearity', 'mask', 'refpix', 'rscd', 'saturation', 'superbias']

>>> header_to_reftypes(test_header("0.7.0", "NRS_BRIGHTOBJ"))
['area', 'camera', 'collimator', 'dark', 'disperser', 'distortion', 'drizpars', 'extract1d', 'filteroffset', 'fore', 'fpa', 'fringe', 'gain', 'ifufore', 'ifupost', 'ifuslicer', 'ipc', 'linearity', 'mask', 'msa', 'ote', 'pathloss', 'photom', 'readnoise', 'refpix', 'regions', 'rscd', 'saturation', 'specwcs', 'straymask', 'superbias', 'v2v3', 'wavelengthrange']

>>> header_to_reftypes(test_header("0.13.0", "MIR_IMAGE"))
['area', 'camera', 'collimator', 'dark', 'dflat', 'disperser', 'distortion', 'drizpars', 'fflat', 'filteroffset', 'flat', 'fore', 'fpa', 'gain', 'ifufore', 'ifupost', 'ifuslicer', 'linearity', 'mask', 'msa', 'ote', 'persat', 'photom', 'readnoise', 'refpix', 'regions', 'rscd', 'saturation', 'sflat', 'specwcs', 'superbias', 'trapdensity', 'trappars', 'wavelengthrange', 'wfssbkg']


>>> header = test_header("0.13.0", "MIR_LRS-FIXEDSLIT")
>>> header_to_pipelines(header)
['calwebb_detector1.cfg', 'calwebb_nrslamp-spec2.cfg']
>>> header_to_reftypes(header)
['barshadow', 'camera', 'collimator', 'cubepar', 'dark', 'disperser', 'distortion', 'drizpars', 'extract1d', 'filteroffset', 'fore', 'fpa', 'gain', 'ifufore', 'ifupost', 'ifuslicer', 'linearity', 'mask', 'msa', 'msaoper', 'ote', 'persat', 'readnoise', 'refpix', 'regions', 'resol', 'rscd', 'saturation', 'specwcs', 'superbias', 'trapdensity', 'trappars', 'wavecorr', 'wavelengthrange', 'wfssbkg']

>>> header_to_pipelines(test_header("0.7.0", "FGS_DARK"))
['calwebb_dark.cfg', 'skip_2b.cfg']

>>> header_to_pipelines(test_header("0.7.0", "NRS_BRIGHTOBJ"))
['calwebb_sloper.cfg', 'calwebb_spec2.cfg']

>>> header = test_header("0.13.0", "MIR_IMAGE", tsovisit="F")
>>> header_to_pipelines(header)
['calwebb_detector1.cfg', 'calwebb_image2.cfg']

>>> header = test_header("0.13.0", "MIR_IMAGE", tsovisit="T")
>>> header_to_pipelines(header)
['calwebb_tso1.cfg', 'calwebb_tso-image2.cfg']

>>> header_to_pipelines(test_header("0.7.0", "MIR_IMAGE"))
['calwebb_sloper.cfg', 'calwebb_image2.cfg']

>>> header_to_pipelines(test_header("0.7.0", "MIR_LRS-FIXEDSLIT"))
['calwebb_sloper.cfg', 'calwebb_spec2.cfg']

>>> header = test_header("0.13.8", "MIR_IMAGE", visitype="GENERIC")
>>> _get_pipelines("MIR_IMAGE", "0.13.8", "jwst_0552.pmap")
['calwebb_detector1.cfg', 'calwebb_image2.cfg']
>>> header_to_pipelines(header)
['calwebb_detector1.cfg', 'calwebb_image2.cfg']

>>> header = test_header("0.13.8", "MIR_IMAGE", visitype="PRIME_WFSC_ROUTINE")
>>> header_to_pipelines(header)
['calwebb_detector1.cfg', 'calwebb_wfs-image2.cfg']

>>> _get_missing_calver("0.7.0")
'0.7.0'

>>> s = _get_missing_calver()
>>> isinstance(s, str)
True

>>> _get_missing_context()
'jwst-operational'

>>> _get_missing_context('jwst_0341.pmap')
'jwst_0341.pmap'

>>> os.path.basename(_get_config_refpath("jwst_0552.pmap", "0.7.7"))
'jwst_system_crdscfg_b7.yaml'

>>> os.path.basename(_get_config_refpath("jwst_0552.pmap", "0.9.3"))
'jwst_system_crdscfg_b7.1.3.yaml'

>>> os.path.basename(_get_config_refpath("jwst_0552.pmap", "0.9.7"))
'jwst_system_crdscfg_b7.1.3.yaml'

>>> os.path.basename(_get_config_refpath("jwst_0477.pmap", "0.10.1"))
'jwst_system_crdscfg_b7.2.yaml'

>>> os.path.basename(_get_config_refpath("jwst_0552.pmap", "0.13.8"))
'jwst_system_crdscfg_b7.4.yaml'
"""

import os.path
from collections import defaultdict
import re

# import yaml     DEFERRED

from pkg_resources import parse_version

# --------------------------------------------------------------------------------------

# from jwst import __version__     DEFERRED

# --------------------------------------------------------------------------------------
import crds
from crds.core import log, utils, config
from crds.core.log import srepr
from crds.client import api

# --------------------------------------------------------------------------------------

def test_header(calver, exp_type, tsovisit="F", lamp_state="OFF", visitype="GENERIC"):
    """Create a header-like dictionary from `calver` and `exp_type` to
    support testing header-based functions.
    """
    header = {
        "META.INSTRUMENT.NAME" : "SYSTEM",
        "REFTYPE" : "CRDSCFG",
        "META.CALIBRATION_SOFTWARE_VERSION" : calver,
        "META.EXPOSURE.TYPE" : exp_type,
        "META.VISIT.TSOVISIT" : tsovisit,
        "META.INSTRUMENT.LAMP_STATE" : lamp_state,
        "META.VISIT.TITLE" : visitype,
        }
    return header

# --------------------------------------------------------------------------------------

def _get_missing_calver(cal_ver=None):
    """If `cal_ver` is None, return the calibration software version for
    the installed version of calibration code.  Otherwise return `cal_ver`
    unchanged.
    """
    if cal_ver is None:
        from jwst import __version__ as calver2
        return calver2
    return cal_ver

def _get_missing_context(context=None):
    """Default the context to `jwst-operational` if `context` is None, otherwise
    return context unchanged.
    """
    return "jwst-operational" if context is None else context

# --------------------------------------------------------------------------------------

def header_to_reftypes(header, context=None):
    """Given a dataset `header`,  extract the EXP_TYPE or META.EXPOSURE.TYPE keyword
    from and use it to look up the reftypes required to process it.

    Return a list of reftype names.
    """
    with log.warn_on_exception("Failed determining reftypes for", log.PP(header)):
        exp_type, cal_ver = _header_to_exptype_calver(header)
        config_manager = _get_config_manager(context, cal_ver)
        pipelines = header_to_pipelines(header, context)
        reftypes = set()
        for cfg in pipelines:
            steps = config_manager.pipeline_cfgs_to_steps[cfg]
            for step in steps:
                reftypes |= set(config_manager.steps_to_reftypes[step])
        return sorted(list(reftypes))
    return []

# This is potentially an external interface to system data processing (SDP) / the archive pipeline.
def header_to_pipelines(header, context=None):
    """Given a dataset `header`,  extract the EXP_TYPE or META.EXPOSURE.TYPE keyword
    from and use it to look up the pipelines required to process it.

    Return a list of pipeline .cfg names.
    """
    with log.augment_exception("Failed determining exp_type, cal_ver from header", log.PP(header)):
        exp_type, cal_ver = _header_to_exptype_calver(header)
    config_manager = _get_config_manager(context, cal_ver)
    pipelines = _get_pipelines(exp_type, cal_ver, context)   # uncorrected
    if config_manager.pipeline_exceptions:  # correction based on extra non-EXP_TYPE params
        pipelines2 = []
        for cfg in pipelines:
            for param, exceptions in config_manager.pipeline_exceptions.items():
                exceptions = dict(exceptions)
                dont_replace = exceptions.pop("dont_replace", [])
                do_replace = exceptions.pop("do_replace", [])
                default_missing  = exceptions.pop("default_missing", "UNDEFINED")
                paramval = header.get(param.upper(), default_missing)
                for dont in dont_replace:
                    if re.match(config.complete_re(dont), paramval):
                        break
                else:
                    if dont_replace:
                        cfg = exceptions.get(cfg, cfg)
                for do in do_replace:
                    if re.match(config.complete_re(do), paramval):
                        cfg = exceptions.get(cfg, cfg)
            pipelines2.append(cfg)
        pipelines = pipelines2
    log.verbose("Applicable pipelines for", srepr(exp_type), "are", srepr(pipelines))
    return pipelines

def _get_pipelines(exp_type, cal_ver=None, context=None):
    """Given `exp_type` and `cal_ver` and `context`,  locate the appropriate SYSTEM CRDSCFG
    reference file and determine the sequence of pipeline .cfgs required to process that
    exp_type.

    NOTE:  This is an uncorrected result,  config_manager.pipeline_exceptions is used to
    alter this based on other header parameters.
    """
    with log.augment_exception("Failed determining required pipeline .cfgs for",
                               "EXP_TYPE", srepr(exp_type)):
        config_manager = _get_config_manager(context, cal_ver)
        return config_manager.exptype_to_pipelines(exp_type)

def reftype_to_pipelines(reftype, cal_ver=None, context=None):
    """Given `exp_type` and `cal_ver` and `context`,  locate the appropriate SYSTEM CRDSCFG
    reference file and determine the sequence of pipeline .cfgs required to process that
    exp_type.
    """
    with log.augment_exception("Failed determining required pipeline .cfgs for",
                               "REFTYPE", srepr(reftype)):
        config_manager = _get_config_manager(context, cal_ver)
        return config_manager.reftype_to_pipelines(reftype)

def _header_to_exptype_calver(header):
    """Given dataset `header`,  return the EXP_TYPE and CAL_VER values."""
    cal_ver = header.get("META.CALIBRATION_SOFTWARE_VERSION", header.get("CAL_VER"))
    cal_ver = _get_missing_calver(cal_ver)
    exp_type = header.get("META.EXPOSURE.TYPE",  header.get("EXP_TYPE", "UNDEFINED"))
    return exp_type, cal_ver

@utils.cached  # for caching,  pars must be immutable, ideally simple
def _get_config_manager(context, cal_ver):
    """Given `context` and calibration s/w version `cal_ver`,  identify the appropriate
    SYSTEM CRDSCFG reference file and create a CrdsCfgManager from it.
    """
    refpath = _get_config_refpath(context, cal_ver)
    return _load_refpath(context, refpath)

def _load_refpath(context, refpath):
    """Given `context` and SYSTEM CRDSCFG reference at `refpath`,  construct a CrdsCfgManager."""
    import yaml
    with open(refpath) as opened:
        crdscfg =  yaml.safe_load(opened)
    return CrdsCfgManager(context, refpath, crdscfg)

# --------------------------------------------------------------------------------------

HERE = os.path.dirname(__file__) or "."

REFPATHS = [
    ('0.7.7', "jwst_system_crdscfg_b7.yaml"),
    ('0.9.0', "jwst_system_crdscfg_b7.1.yaml"),
    ('0.9.1', "jwst_system_crdscfg_b7.1.1.yaml"),
    ('0.9.3', "jwst_system_crdscfg_b7.1.3.yaml"),
    ('0.10.0', "jwst_system_crdscfg_b7.2.yaml"),
    ('0.13.0', "jwst_system_crdscfg_b7.3.yaml"),
    ('0.13.8', "jwst_system_crdscfg_b7.4.yaml"),
    ('0.16.0', "jwst_system_crdscfg_b7.5.yaml"),
    ('0.17.0', "jwst_system_crdscfg_b7.6.yaml"),
    ('1.1.0', "jwst_system_crdscfg_b7.8.yaml"),
    ('999.0.0', "jwst_system_crdscfg_b7.8.yaml"),   # latest backstop
]

def _get_config_refpath(context, cal_ver):
    """Given CRDS `context` and calibration s/w version `cal_ver`,  identify the applicable
    SYSTEM CRDSCFG reference file, cache it, and return the file path.
    """
    context = _get_missing_context(context)
    cal_ver = _get_missing_calver(cal_ver)
    i = 0
    while (i < len(REFPATHS)-1 and not _versions_lt(cal_ver, REFPATHS[i+1][0])):
        i += 1
    refpath = os.path.join(HERE, REFPATHS[i][1])
    try:  # Use a normal try/except because exceptions are expected.
        header = {
            "META.INSTRUMENT.NAME" : "SYSTEM",
            "META.CALIBRATION_SOFTWARE_VERSION": cal_ver
        }
        pmap = crds.get_symbolic_mapping(context)
        imap = pmap.get_imap("system")
        rmapping = imap.get_rmap("crdscfg")
        ref = rmapping.get_best_ref(header)
        refpath = rmapping.locate_file(ref)
        api.dump_references(context, [ref])
    except Exception:
        log.verbose_warning(
            "Failed locating SYSTEM CRDSCFG reference",
            "under context", repr(context),
            "and cal_ver", repr(cal_ver) + ".   Using built-in references.")
    log.verbose("Using", srepr(os.path.basename(refpath)),
                "to determine applicable default reftypes for", srepr(cal_ver))
    return refpath

def _versions_lt(v1, v2):
    """Compare two semantic version numbers and account for issues like '10' < '9'.
    Return True IFF `v1` < `v2`.

    >>> _versions_gte("0.7.7", "0.7.7")
    True
    >>> _versions_gte("0.7.8", "0.7.7")
    True
    >>> _versions_gte("0.7.6", "0.7.7")
    False
    >>> _versions_gte("0.7.7", "0.7.8")
    False
    >>> _versions_gte("0.7.7", "0.7.6")
    True
    >>> _versions_gte("0.10.1", "0.7.7")
    True
    >>> _versions_gte("0.10.1", "0.10.0")
    True
    >>> _versions_gte("0.10.0", "0.10.1")
    False
    >>> _versions_gte("0.10.1", "0.10.1")
    True
    >>> _versions_gte("0.10.1", "0.10.1a1")
    True
    >>> _versions_gte("0.10.2", "0.10.1a1")
    True
    >>> _versions_gte("0.10.1", "0.11.1a1")
    False
    >>> _versions_gte("0.9.7", "0.10.0")
    False
    >>> _versions_gte("0.9.2", "0.10.1dev20000")
    False
    >>> _versions_gte("0.99.7", "0.10.1dev20000")
    True
    >>> _versions_gte("0.10.1", "0.10.1dev20000")
    True
    """
    return _reduce_ver(v1) < _reduce_ver(v2)

def _versions_gte(v1, v2):
    """"""
    return _reduce_ver(v1) >= _reduce_ver(v2)

def _reduce_ver(ver):
    """
    >>> _reduce_ver("1")
    (1, 0, 0)
    >>> _reduce_ver("1.2")
    (1, 2, 0)
    >>> _reduce_ver("1.2.3")
    (1, 2, 3)
    >>> _reduce_ver("1.2.3.4")
    (1, 2, 3)
    """
    ver = parse_version(ver).base_version
    parts = ver.split(".")
    while len(parts) < 3:
        parts.append("0")
    return tuple(map(int, parts))[:3]

class CrdsCfgManager:
    """The CrdsCfgManager handles using SYSTEM CRDSCFG information to compute things."""
    def __init__(self, context, refpath, crdscfg):
        self._context = context
        self._refpath = refpath
        self._crdscfg = utils.Struct(crdscfg)
        self.pipeline_exceptions = self._crdscfg.get("pipeline_exceptions", {})
        self.pipeline_cfgs_to_steps = self._crdscfg.get("pipeline_cfgs_to_steps", {})
        self.steps_to_reftypes = self._crdscfg.get("steps_to_reftypes", {})

    def exptype_to_reftypes(self, exp_type):
        """For a given EXP_TYPE string, return a list of reftypes needed to process that
        EXP_TYPE through the data levels appropriate for that EXP_TYPE.

        Return [reftypes, ... ]
        """
        reftypes = self._crdscfg.exptypes_to_reftypes[exp_type]
        log.verbose("Applicable reftypes for", srepr(exp_type),
                    "determined by", srepr(os.path.basename(self._refpath)),
                    "are", srepr(reftypes))
        return reftypes

    def exptype_to_pipelines(self, exp_type):
        """For a given EXP_TYPE string, return a list of pipeline .cfg's needed to
        process that EXP_TYPE through the appropriate data levels.

        Return [.cfg's, ... ]
        """
        pipelines = self._crdscfg.exptypes_to_pipelines[exp_type]
        return pipelines

    def reftype_to_pipelines(self, reftype):
        pipelines = []
        reftypes_to_steps = invert_list_mapping(self._crdscfg.steps_to_reftypes)
        steps_to_pipelines = invert_list_mapping(self._crdscfg.pipeline_cfgs_to_steps)
        for step in reftypes_to_steps[reftype]:
            pipelines += steps_to_pipelines[step]
        return list(sorted(set(pipelines)))

def invert_list_mapping(mapping):
    """Invert a dictionary of lists into another dictionary of lists such that each
    element of each original list is a key somewhere in the inversion, and each key is
    an element of at least one list in the inversion.
    """
    inverted = defaultdict(set)
    for key, values in mapping.items():
        for value in values:
            inverted[value].add(key)
    return { key:list(sorted(values))
             for (key,values) in inverted.items() }

def scan_exp_type_coverage():
    """Verify that there is some get_reftypes() response for all available exp_types."""
    from . import schema as crds_schema
    exp_types = crds_schema.get_exptypes()
    for exp_type in exp_types:
        if exp_type in ["ANY","N/A"]:
            continue
        with log.warn_on_exception("failed determining reftypes for", repr(exp_type)):
            reftypes = _get_reftypes(exp_type)
            log.verbose("Reftypes for", repr(exp_type), "=", repr(reftypes))

def _get_reftypes(exp_type, cal_ver=None, context=None):
    """Given `exp_type` and `cal_ver` and `context`,  locate the appropriate SYSTEM CRDSCFG
    reference file and determine the reference types required to process every pipeline Step
    nominally associated with that exp_type.
    """
    with log.warn_on_exception("Failed determining required reftypes from",
                               "EXP_TYPE", srepr(exp_type)):
        config_manager = _get_config_manager(context, cal_ver)
        return config_manager.exptype_to_reftypes(exp_type)
    return []

def test():
    """Run the module doctests."""
    import doctest
    from crds.jwst import pipeline
    return doctest.testmod(pipeline)

# --------------------------------------------------------------------------------------

if __name__ == "__main__":
    print(test())



"""This module implements a get_data_level(reftype) function that returns the
first processing level at which the requested reftype is used.
"""
import json

import crds

def get_data_level(
    reftype, calibration_version, context=None, extras=None, observatory="jwst", ignore_cache=False):
    """Based on `context` and `calibration_version` (nominally CAL_VER), return the
    minimum data level associated with reference type `reftype`.  Nominally
    this means fetching and interpreting the SYSTEM DATALVL reference file so
    utilizing this function requires operating in an environment with an
    up-to-date and accessible CRDS cache.
    """
    header = {
        "META.INSTRUMENT.NAME":"SYSTEM",
        "INSTRUME":"SYSTEM",
        "INSTRUMENT":"SYSTEM",
        "META.CALIBRATION_SOFTWARE_VERSION":calibration_version,
        "CAL_VER":calibration_version,
        "VERSION":calibration_version,
        }
    header.update(extras or {})
    bestrefs = crds.getreferences(
        header, context=context, reftypes=["datalvl"], observatory=observatory, ignore_cache=False)
    reference_path = bestrefs["datalvl"]
    type_to_level = get_type_to_level(reference_path)
    return type_to_level[reftype]

def get_type_to_level(refpath):
    """Given SYSTEM DATALVL reference file path `refpath`,  load it,
    and return the type_to_level mapping implied by the cal_levels field.
    """
    loaded = json.load(open(refpath))
    return invert_cal_levels(loaded["cal_levels"])

def invert_cal_levels(cal_levels):
    """Invert and return the cal_levels mapping.

    >>> cal_levels = {
    ...    "2A": [
    ...        "dark",
    ...        "gain",
    ...        "ipc",
    ...    ],
    ...    "2B": [
    ...        "area",
    ...        "camera",
    ...        "collimator",
    ...    ],
    ...    "3": [
    ...        "distortion",
    ...        "drizpars",
    ...    ]
    ... }

    >>> inverted = invert_cal_levels(cal_levels)
    >>> sorted(inverted.items())
    [('area', '2B'), ('camera', '2B'), ('collimator', '2B'), ('dark', '2A'), ('distortion', '3'), ('drizpars', '3'), ('gain', '2A'), ('ipc', '2A')]

    >>> tricky_levels = {
    ...    "1A" : [ "distortion"],
    ...    "2A" : [ "distortion"],
    ...    }
    >>> sorted(invert_cal_levels(tricky_levels).items())
    [('distortion', '1A')]
    """
    type_to_level = dict()
    for level in cal_levels:
        for type in cal_levels[level]:
            if type in type_to_level:
                existing = type_to_level[type]
                if level < existing:
                    type_to_level[type] = level
            else:
                type_to_level[type] = level
    return type_to_level

def test():
    import doctest
    from crds.misc import datalvl
    doctest.testmod(datalvl)
